/*******************************************************************************
				Neoinput
				========

		Copyright (C) 2010 Paul Kocialkowski

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include <linux/input.h>
#include <linux/uinput.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "server.h"
#include "input_read.h"
#include "input.h"

#include "input_write.h"

int input=0;
char *input_device=NULL;
float speed_factor=1;

char *uinput_devices[3]={"/dev/input/uinput/", "/dev/uinput", "/dev/misc/uinput"};

void input_open_write(int input_type)
{
	struct uinput_user_dev uinput_device;
	int i;

	/* Use uinput if nothing specified */
	if(input_device == NULL)
	{
		/* Try each uinput device */
		for(i=0 ; i < 2 ; i++)
		{
			input=open(uinput_devices[i], O_WRONLY | O_NONBLOCK);

			if(input < 0)
				continue;
			else
				break;
		}
	}
	else
		input=open(input_device, O_WRONLY);

	if(input < 0)
	{
		if(input_device == NULL)
			printf("Error while opening uinput.\n");
		else
			printf("Error while opening %s.\n", input_device);
		exit(1);
	}
	printf("input is %s\n", uinput_devices[i]);

	switch(input_type)
	{
		case INPUT_TYPE_TOUCH:
		case INPUT_TYPE_TOUCH_ROTATE:
		case INPUT_TYPE_ACCEL:
			ioctl(input, UI_SET_EVBIT, EV_KEY);
			ioctl(input, UI_SET_KEYBIT, BTN_LEFT);
			ioctl(input, UI_SET_KEYBIT, BTN_RIGHT);

			ioctl(input, UI_SET_EVBIT, EV_REL); 
			ioctl(input, UI_SET_RELBIT, REL_X);
			ioctl(input, UI_SET_RELBIT, REL_Y);

			ioctl(input, UI_SET_EVBIT, EV_SYN);

			if(input_type == INPUT_TYPE_ACCEL)
			{
				ioctl(input, UI_SET_KEYBIT, KEY_UP);
				ioctl(input, UI_SET_KEYBIT, KEY_DOWN);
				ioctl(input, UI_SET_KEYBIT, KEY_LEFT);
				ioctl(input, UI_SET_KEYBIT, KEY_RIGHT);
				ioctl(input, UI_SET_KEYBIT, KEY_RIGHTSHIFT);
				ioctl(input, UI_SET_KEYBIT, KEY_SPACE);
			}

			if(input_device == NULL)
			{
				memset(&uinput_device, 0, sizeof(uinput_device));
				snprintf(uinput_device.name, UINPUT_MAX_NAME_SIZE, "uinput-sample");
				uinput_device.id.bustype = BUS_USB;
				uinput_device.id.vendor  = 0x1;
				uinput_device.id.product = 0x1;
				uinput_device.id.version = 1;

				write(input, &uinput_device, sizeof(uinput_device));

				ioctl(input, UI_DEV_CREATE);
			}
		break;
	}
}

void input_event_get(int input_type)
{
	if(!sock_client_fd)
	{
		printf("Socket error.");
		exit(1);
	}

	struct input_event_light event_light;
	struct input_event event;

	__s32 event_light_value_old_0=0;
	__s32 event_light_value_old_1=0;

	int up_lock=0;
	int down_lock=0;
	int right_lock=0;
	int left_lock=0;
	int shift_lock=0;

	int pause_lock=1;

	switch(input_type)
	{
		case INPUT_TYPE_TOUCH:
		case INPUT_TYPE_TOUCH_ROTATE:
			while(read(sock_client_fd, &event_light, sizeof(struct input_event_light)))
			{
				if(event_light.type == 3)
				{
					event.type=EV_REL;
		
					if(event_light.code == 0)
					{
						if(!event_light_value_old_0)
							event_light_value_old_0=event_light.value;

						if(input_type == INPUT_TYPE_TOUCH_ROTATE)
							event.code=REL_X;
						else
							event.code=REL_Y;

						event.value=event_light_value_old_0 - event_light.value;
						event.value*=speed_factor;

						event_light_value_old_0=event_light.value;

						input_event_write(event);

					}
					else if(event_light.code == 1)
					{
						if(!event_light_value_old_1)
							event_light_value_old_1=event_light.value;

						event.value=event_light_value_old_1 - event_light.value;

						if(input_type == INPUT_TYPE_TOUCH_ROTATE)
							event.code=REL_Y;
						else
						{
							event.code=REL_X;
							event.value*=-1;

						}

						event.value*=speed_factor;

						event_light_value_old_1=event_light.value;

						input_event_write(event);
					}
					else if(event_light.code == 24 && event_light.value == 0)
					{
						event_light_value_old_0=0;
						event_light_value_old_1=0;
					}
				}
				if(event_light.type == 1 && event_light.code == 169)
				{
					if(event_light.value == 1)
					{
						event.type=EV_KEY;
						event.code = BTN_LEFT;
						event.value = 1;

						input_event_write(event);
					}
					else if(event_light.value == 0)
					{
						event.type=EV_KEY;
						event.code = BTN_LEFT;
						event.value = 0;

						input_event_write(event);
					}
				}
				if(event_light.type == 1 && event_light.code == 116)
				{
					if(event_light.value == 1)
					{
						event.type=EV_KEY;
						event.code = BTN_RIGHT;
						event.value = 1;

						input_event_write(event);
					}
					else if(event_light.value == 0)
					{
						event.type=EV_KEY;
						event.code = BTN_RIGHT;
						event.value = 0;

						input_event_write(event);
					}
				}
				if(event_light.type == 0 && event_light.code == 0 && event_light.value == 0)
				{
					event.type = EV_SYN;
					event.code = SYN_REPORT;
					event.value = 0;

					input_event_write(event);
				}
			}
		break;
		case INPUT_TYPE_ACCEL:
			while(read(sock_client_fd, &event_light, sizeof(struct input_event_light)))
			{
				if(event_light.type == 1 && event_light.code == 169 && event_light.value == 1)
				{
					if(!pause_lock)
						pause_lock=1;
					else
						pause_lock=0;
				}

				if(event_light.type == 1 && event_light.code == 116)
				{
					if(event_light.value == 1)
					{
						event.type=EV_KEY;
						event.code = KEY_SPACE;
						event.value = 1;

						input_event_write(event);
					}
					else if(event_light.value == 0)
					{
						event.type=EV_KEY;
						event.code = KEY_SPACE;
						event.value = 0;

						input_event_write(event);
					}

					event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
					input_event_write(event);
				}

				if(pause_lock)
				{
					event.type=EV_KEY;

					if(up_lock)
					{
						event.code=KEY_UP;
						event.value=0;
						input_event_write(event);

						event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
						input_event_write(event);

						up_lock=0;
					}
					if(down_lock)
					{
						event.code=KEY_DOWN;
						event.value=0;
						input_event_write(event);

						event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
						input_event_write(event);

						down_lock=0;
					}
					if(left_lock)
					{
						event.code=KEY_LEFT;
						event.value=0;
						input_event_write(event);

						event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
						input_event_write(event);

						left_lock=0;
					}
					if(right_lock)
					{
						event.code=KEY_RIGHT;
						event.value=0;
						input_event_write(event);

						event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
						input_event_write(event);

						right_lock=0;
					}
					if(shift_lock)
					{
						event.code=KEY_RIGHTSHIFT;
						event.value=0;
						input_event_write(event);

						event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
						input_event_write(event);

						shift_lock=0;
					}

					continue;
				}

				if(event_light.type == 3)
				{
					event.type=EV_KEY;

					/*
					1000 à -1000
					1000 -> penché en avant
					0 -> plat
					-1000 -> penché en arrière

					---> après -750, on considère qu'on freine
					---> entre -250 et -750 on fait rien
					---> entre 1000 et -250, on accel
					*/

					if(event_light.code == 0)
					{
						if(event_light.value > -250 || event_light.value < -750)
						{
							if(event_light.value > -250)
							{
								if(!up_lock)
								{
									event.code=KEY_UP;
									event.value=1;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);

									up_lock=1;
									down_lock=0;
								}
								else
								{
									event.code=KEY_UP;
									event.value=2;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);
								}
							}
							else if(up_lock)
							{
								event.code=KEY_UP;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								up_lock=0;
							}

							if(event_light.value < -750)
							{
								if(!down_lock)
								{
									event.code=KEY_DOWN;
									event.value=1;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);

									up_lock=0;
									down_lock=1;
								}
								else
								{
									event.code=KEY_DOWN;
									event.value=2;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);
								}
							}
							else if(down_lock)
							{
								event.code=KEY_DOWN;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								down_lock=0;
							}
						}
						else
						{
							if(up_lock)
							{
								event.code=KEY_UP;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								up_lock=0;
							}
							if(down_lock)
							{
								event.code=KEY_DOWN;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								down_lock=0;
							}
						}
					}

					/*
					75 -> -75  : rien
					0 -> 750   : petite inclinaison gauche -> KEY_RIGHTSHIFT
					0 -> 1000  : grosse inclinaison gauche
					0 -> -750  : petite inclinaison droite -> KEY_RIGHTSHIFT
					0 -> -1000 : grosse inclinaison droite

					0 - 750    : par accoups
					750 - * : en continu
					*/

					if(event_light.code == 1)
					{
						if(event_light.value > 250 || event_light.value < -250)
						{
							if(event_light.value > 0)
							{
								if(!left_lock)
								{
									event.code=KEY_LEFT;
									event.value=1;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);

									right_lock=0;
									left_lock=1;
								}
								else
								{
									event.code=KEY_LEFT;
									event.value=2;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);
								}
							}
							else if(left_lock)
							{
								event.code=KEY_LEFT;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								left_lock=0;
							}
					
							if(event_light.value < 0)
							{
								if(!right_lock)
								{
									event.code=KEY_RIGHT;
									event.value=1;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);

									right_lock=1;
									left_lock=0;
								}
								else
								{
									event.code=KEY_RIGHT;
									event.value=2;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);
								}
							}
							else if(right_lock)
							{
								event.code=KEY_RIGHT;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								right_lock=0;
							}

	
							if((event_light.value > 0 && event_light.value < 750) || (event_light.value < 0 && event_light.value > -750))
							{
								if(!shift_lock)
								{
									event.code=KEY_RIGHTSHIFT;
									event.value=1;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);

									shift_lock=1;
								}
								else
								{
									event.code=KEY_RIGHTSHIFT;
									event.value=2;
									input_event_write(event);

									event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
									input_event_write(event);
								}
							}
							else if(shift_lock)
							{
								event.code=KEY_RIGHTSHIFT;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								shift_lock=0;
							}
						}
						else
						{
							if(left_lock)
							{
								event.code=KEY_LEFT;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								left_lock=0;
							}
							if(right_lock)
							{
								event.code=KEY_RIGHT;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								right_lock=0;
							}
							if(shift_lock)
							{
								event.code=KEY_RIGHTSHIFT;
								event.value=0;
								input_event_write(event);

								event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
								input_event_write(event);

								shift_lock=0;
							}

						}
					}
				}
			}

			if(up_lock)
			{
				event.code=KEY_UP;
				event.value=0;
				input_event_write(event);

				event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
				input_event_write(event);

				up_lock=0;
			}
			if(down_lock)
			{
				event.code=KEY_DOWN;
				event.value=0;
				input_event_write(event);

				event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
				input_event_write(event);

				down_lock=0;
			}
			if(left_lock)
			{
				event.code=KEY_LEFT;
				event.value=0;
				input_event_write(event);

				event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
				input_event_write(event);

				left_lock=0;
			}
			if(right_lock)
			{
				event.code=KEY_RIGHT;
				event.value=0;
				input_event_write(event);

				event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
				input_event_write(event);

				right_lock=0;
			}
			if(shift_lock)
			{
				event.code=KEY_RIGHTSHIFT;
				event.value=0;
				input_event_write(event);

				event.type = EV_SYN; event.code = SYN_REPORT; event.value = 0;
				input_event_write(event);

				shift_lock=0;
			}
		break;
	}

}

void input_event_write(struct input_event data)
{
	if(!write(input, &data, sizeof(struct input_event)))
	{
		perror("Error in write()");
		exit(1);
	}
}

void input_close(void)
{
	if(input_device == NULL)
		ioctl(input, UI_DEV_DESTROY);
	close(input);
}
